/*-
 * Copyright (C)2020..2021 @BABOLO http://www.babolo.ru/
 * PKG = babolo-libmake
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2020..2021 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: pout.c,v 1.2 2021/08/22 15:57:57 babolo Exp $\n"

#define Bpars_COMPAT  4
#define MULAR_COMPAT  0
#define MIFE_COMPAT   5
#define BLIN_COMPAT   VMAJOR
#define BLIN_INTERNAL

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include "../BLINflag.h"
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "blinread.h"

int
BLINr_rp(BLINr_cfg *cfg, BLIN_trace *twind) {
#   define isJX(EXCODE, FORMAT, ...) \
        do { ifBLIN_QX0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define isJW(EXCODE, FORMAT, ...) \
        do { ifBLIN_QW0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define blin_internal_flags (cfg->flags & BLIN_MASK)
    BLINr_fraze      *curfraze;                /*                           */
    u_int64_t         fstime;                  /*    tracfile                   */
    time_t            hhtime;                  /*    tm                    */
    struct tm        *htime = NULL;            /*       ISO */
    int               ex = EX_OK;              /*                                */
    u_int32_t         p;                       /* ޣ,              */

    qbuf.time1 = qbuf.time2 = fstime = twind->time;
    for (u_int8_t truvar, termvar; ; twind = BLINr_next_wind(cfg)) {  /*****/
        if  (!twind) isJW(EX_IOERR, "tracefile mife_window %"BLIN_U, cfg->tfile->ofpoint);
        ifBLIN_QX3( "twind={%u %u %u %"BLIN_O"u %u %u %u %u %u %.*s}"
                  , twind->len
                  , twind->level
                  , twind->pid
                  , twind->time
                  , twind->erno
                  , twind->line
                  , twind->snprog
                  , twind->snfile
                  , twind->snfunc
                  , (int)(twind->len - offsetof(BLIN_trace, string))
                  , twind->string
                  )
        ;
        for (curfraze = cfg->filt; curfraze; curfraze = curfraze->next) {
            ifBLIN_QX3("curfraze3 %X", curfraze);
            if  (curfraze->flags & BLINREAD_X) {
                truvar = BLINREAD_M;
                termvar = 0;
            }
            if  (curfraze->time.from == curfraze->time.to) {
                if  (!(termvar & BLINREAD_A)) truvar |= BLINREAD_A;
            } else {
                truvar &= ~BLINREAD_A;
                termvar |= BLINREAD_A;
                p = 0;
                for ( BLINr_time *i = mular_getix(cfg->time, p++)
                    ; p <= curfraze->time.to
                    ; i = mular_getix(cfg->time, p++)
                    ) {
                    if  (!i) isJW(1, "check time mular_getix");
                    if  ( i->nrng
                        ? (i->from + (i->abs ? 0 : fstime)) == twind->time
                        :    ((i->from + (i->abs ? 0 : fstime)) <= twind->time)
                          && (twind->time <= (i->to + (i->abs ? 0 : fstime)))
                        ) {
                        truvar |= BLINREAD_A;
                        break;
            }   }   }
            if  (curfraze->pid.from == curfraze->pid.to) {
                if  (!(termvar & BLINREAD_I)) truvar |= BLINREAD_I;
            } else {
                truvar &= ~BLINREAD_I;
                termvar |= BLINREAD_I;
                p = 0;
                for ( u_int32_t *i = mular_getix(cfg->pid, p++)
                    ; p <= curfraze->pid.to
                    ; i = mular_getix(cfg->pid, p++)
                    )
                {
                    if  (!i) isJW(1, "check pid mular_getix");
                    if  (*i == twind->pid) {
                        truvar |= BLINREAD_I;
                        break;
            }   }   }
            if  (curfraze->prog.from == curfraze->prog.to) {
                if  (!(termvar & BLINREAD_R)) truvar |= BLINREAD_R;
            } else {
                truvar &= ~BLINREAD_R;
                termvar |= BLINREAD_R;
                p = 0;
                for ( const char **i = mular_getix(cfg->prog, p++)
                    ; p <= curfraze->prog.to
                    ; i = mular_getix(cfg->prog, p++)
                    ) {
                    if  (!i) isJW(1, "check prog mular_getix");
                    if  (  (strlen(*i) == twind->snprog)
                        && !strncmp(*i, &twind->string[0], twind->snprog)
                        ) {
                        truvar |= BLINREAD_R;
                        break;
            }   }   }
            if  (curfraze->file.from == curfraze->file.to) {
                if  (!(termvar & BLINREAD_S)) truvar |= BLINREAD_S;
            } else {
                truvar &= ~BLINREAD_S;
                termvar |= BLINREAD_S;
                ifBLIN_QX3( "%u %.*s"
                          , twind->snfile
                          , (int) twind->snfile
                          , &twind->string[twind->snprog]
                          )
                ;
                p = 0;
                for ( const char **i = mular_getix(cfg->file, p++)
                    ; p <= curfraze->file.to
                    ; i = mular_getix(cfg->file, p++)
                    ) {
                    if  (!i) isJW(1, "check file mular_getix");
                    ifBLIN_QX3("check file mular %zd %s", strlen(*i), *i);
                    if  (  (strlen(*i) == twind->snfile)
                        && !(strncmp(*i, &twind->string[twind->snprog], twind->snfile))
                        ) {
                        truvar |= BLINREAD_S;
                        break;
            }   }   }
            if  (curfraze->func.from == curfraze->func.to) {
                if  (!(termvar & BLINREAD_U)) truvar |= BLINREAD_U;
            } else {
                truvar &= ~BLINREAD_U;
                termvar |= BLINREAD_U;
                p = 0;
                for ( const char **i = mular_getix(cfg->func, p++)
                    ; p <= curfraze->func.to
                    ; i = mular_getix(cfg->func, p++)
                    ) {
                    if  (!i) isJW(1, "check func mular_getix");
                    if  (  (strlen(*i) == twind->snfunc)
                        && !strncmp( *i
                                   , &twind->string[twind->snprog + twind->snfile]
                                   , twind->snfunc
                        )          ) {
                        truvar |= BLINREAD_U;
                        break;
            }   }   }
            if  (curfraze->line.from == curfraze->line.to) {
                if  (!(termvar & BLINREAD_N)) truvar |= BLINREAD_N;
            } else {
                truvar &= ~BLINREAD_N;
                termvar |= BLINREAD_N;
                p = 0;
                for ( BLINr_line *i = mular_getix(cfg->line, p++)
                    ; p <= curfraze->line.to
                    ; i = mular_getix(cfg->line, p++)
                    ) {
                    if  (!i) isJW(1, "check line mular_getix");
                    if  ( i->nrng
                        ? i->lnfr == twind->line
                        : (i->lnfr <= twind->line) && (twind->line <= i->lnto)
                        ) {
                        truvar |= BLINREAD_N;
                        break;
            }   }   }
            if  (!curfraze->levfl) {
                if  (!(termvar & BLINREAD_L)) truvar |= BLINREAD_L;
            } else {
                truvar &= ~BLINREAD_L;
                termvar |= BLINREAD_L;
                if  (curfraze->level & (1U << twind->level)) truvar |= BLINREAD_L;
            }
            truvar ^= curfraze->flags & BLINREAD_M;
            if  (truvar == BLINREAD_M) break;
        }
        if  (truvar == BLINREAD_M) {
            if  (cfg->flags & BLINREAD_P_I) printf("%u ", twind->pid);
            if  (cfg->flags & BLINREAD_P_T) {
                printf( "%04"BLIN_O"u.%09u "
                      , (twind->time - fstime) / BLINr_E9
                      , (u_int32_t)((twind->time - fstime) % BLINr_E9)
                      )
                ;
            } else if  (cfg->flags & BLINREAD_P_D) {
                hhtime = twind->time / BLINr_E9;
                htime = localtime((const time_t*)&hhtime);
                printf( "%u-%02u-%02u %02u:%02u:%02u.%09u "
                      , htime->tm_year + 1900
                      , htime->tm_mon + 1
                      , htime->tm_mday
                      , htime->tm_hour
                      , htime->tm_min
                      , htime->tm_sec
                      , (u_int32_t)(twind->time % BLINr_E9)
                      )
                ;
            }
            if  (cfg->flags & BLINREAD_P_L) printf("%u ", twind->level);
            if  (cfg->flags & BLINREAD_P_E) printf("%u ", twind->erno);
            if  (cfg->flags & BLINREAD_P_R) printf("%.*s:", twind->snprog, twind->string);
            if  (cfg->flags & BLINREAD_P_S) {
                printf("%.*s:", twind->snfile, &twind->string[twind->snprog]);
            }
            if  (cfg->flags & BLINREAD_P_U) {
                printf( "%.*s:"
                      , twind->snfunc
                      , &twind->string[twind->snprog + twind->snfile]
                      )
                ;
            }
            if  (cfg->flags & BLINREAD_P_N) printf("%u:", twind->line);
            for ( p = 0
                ; twind->string[BLINr_slen(twind) - 1 - p] == '\n'
                ; p++
                ) {}
            printf( " %.*s\n"
                  , (int)( BLINr_slen(twind) - twind->snprog - twind->snfile - twind->snfunc - p)
                  , &twind->string[twind->snprog + twind->snfile + twind->snfunc]
                  )
            ;
        }
        if  (mife_ctlsize(cfg->tfile) == twind->len) break;
    }
out:
    return(ex);
#   undef isJX
#   undef isJW
#   undef blin_internal_flags
}
